/*
 * TODO:
 * 	- signal twice to have original (especially CTRL-C in CTRL-Z)
 * 	- see other hash functions
 * 	- compteur_t clean-up
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <stdint.h>
#include <time.h>

#define DEFAULT_SIZE 256		/* buffer for file reading */


/*
 *******************************************************************************
 * Types
 */
typedef enum {
	E_Vide,
	E_Pion,	
	E_Interdit
} etat_t;

typedef enum {
	S_NULL = 0,
	S_Nord = 1 << 0,
	S_Est  = 1 << 1,
	S_Ouest= 1 << 2,
	S_Sud  = 1 << 3
} sens_t;

#define Name(sens) \
	(sens == S_Nord ? "Nord" :  \
	 (sens == S_Sud ? "Sud" : \
	  (sens == S_Est ? "Est" :  \
	   (sens == S_Ouest ? "Ouest" : \
	    "Invalid")))) 

typedef struct {
	unsigned int x;
	unsigned int y;
	sens_t s;
} mouvement_t;

typedef struct {
	unsigned int nb_col;
	unsigned int nb_lig;
	int x_fin;
	int y_fin;
	unsigned int nb_pion;
	etat_t *grille;
} plateau_t;

struct cell {
	struct cell *next;
	int hit;
	uint8_t sign[1];
};

struct hash_item {
	int hit;
	uint8_t *sign;
};

struct hash {
	uint16_t nb_items;
	uint16_t size;
	struct hash_item *items;
};


typedef  uint64_t compteur_t;
struct stat_item {
	compteur_t value;
	unsigned int key;
};


typedef struct {
	mouvement_t *liste;	/* liste des mouvements en cours */
	unsigned int nb_pion;	/* nb de pions totaux */
	plateau_t *plateau;	/* plateau en cours' */
	unsigned int *random;	/* unused */
	unsigned int max_hash_size;	/* max d'items pour une cle */
	uint32_t hash_size;	/* taille du hash */
	struct hash *hash;	/* hash des 'tableaux' essayes */
	unsigned int sign_sz;		/* taille de la signature */
	/* symteries possibles */
	int V_sym;
	int H_sym;
	int S_sym;
	int B_sym;
	/* pour les stats. Initialise une fois car malloc n'est pas 'reentrant'
	 * (avec signal) */
	struct stat_item *array;
	int stats;
} solution_t;


#define REMEMBER(pass,nb_pion) ((pass % 2) && pass > 3 && pass < (int)nb_pion-2)

#define MAX(x,y) ((x)>(y)?(x):(y))

#if 0 
#define return_error(rc, fmt, ...) \
	do { \
		fprintf(stderr, "Error:" fmt "\n",##__VA_ARGS__); \
		return rc; \
	} while (0)

#define info(fmt, ...) \
	fprintf(stderr, fmt "\n",##__VA_ARGS__)

#define debug(fmt, ...) \
	fprintf(stderr, "Dbg:" fmt "\n", ##__VA_ARGS__) 

#define MALLOC(num,var,rc,fmt,...) \
	do { \
		if (!(var = malloc(num*sizeof(*var)))) \
			return_error2(rc, "Cannot allocate %zd*%zu:" fmt, \
				     (size_t)num,sizeof(*var),##__VA_ARGS__); \
	} while (0)

#define REALLOC(num,var,rc,fmt,...) \
	do { \
		if (!(var = realloc(var,num*sizeof(*var)))) \
			return_error2(rc, "Cannot re-allocate %zd*%zu:" fmt, \
				     (size_t)num,sizeof(*var),##__VA_ARGS__); \
	} while (0)

#define CALLOC(num,var,rc,fmt,...) \
	do { \
		if (!(var = calloc(num,sizeof(*var)))) \
			return_error2(rc, "Cannot callocate %zd*%zu:" fmt, \
				     (size_t)num,sizeof(*var),##__VA_ARGS__); \
	} while (0)

#else
#define return_error(rc, x) \
	do { \
		fprintf(stderr, "Error: " x "\n"); \
		return rc; \
	} while (0)
#define return_error1(rc, x,y) \
	do { \
		fprintf(stderr, "Error: " x "\n",y); \
		return rc; \
	} while (0)
#define return_error2(rc, x,y,z) \
	do { \
		fprintf(stderr, "Error: " x "\n",y,z); \
		return rc; \
	} while (0)
#define return_error3(rc, x,y,z,aa) \
	do { \
		fprintf(stderr, "Error: " x "\n",y,z,aa); \
		return rc; \
	} while (0)
#define return_error4(rc, x,y,z,aa,bb) \
	do { \
		fprintf(stderr, "Error: " x "\n",y,z,aa,bb); \
		return rc; \
	} while (0)

#define info(x) \
	fprintf(stderr, x "\n")
#define info1(x,y) \
	fprintf(stderr, x "\n",y)
#define info2(x,y,z) \
	fprintf(stderr, x "\n",y,z)
#define info4(x,y,z,aa,bb) \
	fprintf(stderr, x "\n",y,z,aa,bb)


#define MALLOC(num,var,rc, x ) \
	do { \
		if (!(var = malloc(num*sizeof(*var)))) \
			return_error2(rc, "Cannot allocate %d*%u:" x, \
				     (int)num,(unsigned int)sizeof(*var)); \
	} while (0)

#define REALLOC(num,var,rc, x,y,z) \
	do { \
		if (!(var = realloc(var,num*sizeof(*var)))) \
			return_error4(rc, "Cannot re-allocate %d*%u:" x, \
				     (int)num,(unsigned int)sizeof(*var),y,z); \
	} while (0)
#define CALLOC(num,var,rc, x) \
	do { \
		if (!(var = calloc(num,sizeof(*var)))) \
			return_error2(rc, "Cannot callocate %d*%u:" x, \
				     (int)num,(unsigned int)sizeof(*var)); \
	} while (0)
#endif



#define CaseGrille(g, nb_col, i, j) (g[(nb_col*(j))+(i)])
#define Case(p,i,j)       (CaseGrille(p->grille, p->nb_col, i, j))
#define Nord(p,i,j)       (CaseGrille(p->grille, p->nb_col, i, j-1))
#define NordNord(p,i,j)   (CaseGrille(p->grille, p->nb_col, i, j-2))
#define Sud(p,i,j)        (CaseGrille(p->grille, p->nb_col, i, j+1))
#define SudSud(p,i,j)     (CaseGrille(p->grille, p->nb_col, i, j+2))
#define Est(p,i,j)        (CaseGrille(p->grille, p->nb_col, i+1, j))
#define EstEst(p,i,j)     (CaseGrille(p->grille, p->nb_col, i+2, j))
#define Ouest(p,i,j)      (CaseGrille(p->grille, p->nb_col, i-1, j))
#define OuestOuest(p,i,j) (CaseGrille(p->grille, p->nb_col, i-2, j))

static void StatHash(solution_t *s);
static int DisplayPlateau(const plateau_t *p);
static void FreePlateau(plateau_t *p);

static solution_t *g_sol = NULL;

static void (*oldint)(int);
static void (*oldtstp)(int);

static void handler(int sig)
{
	info2("\nCaught signal %d%s",sig,
	     sig == SIGINT?". Hit CTRL-C once more if stuck":"");
	signal(SIGINT, oldint);
	if (g_sol)
		StatHash(g_sol);
	if (sig == SIGINT)
		exit(EXIT_FAILURE);
	signal(SIGINT, handler);
}

/*
 *******************************************************************************
 * Signature/Hash part
 */
static void DisplaySign(uint8_t *sign, size_t sz, uint32_t key)
{
	unsigned int i;
	printf("Hash:(key=%x) ", key);
	for (i = 0; i < sz; ++i)
		printf("%02x", (unsigned int)sign[i]);
	printf("\n");
	return ;
}

#define SIGN(letter,_i,_j) \
static uint8_t *Calc##letter##Sign(solution_t *s, uint32_t *key) \
{ \
	unsigned int i,j; \
	uint8_t x = 0x1 << 7; \
	uint32_t ikey = 0x00; \
	size_t ndx = 0; \
	uint8_t *sign; \
	size_t sz = s->sign_sz; \
	plateau_t *p = s->plateau; \
	*key = 0; \
	CALLOC(sz, sign, NULL, "Signature"); \
	for (j = 0; j < p->nb_lig; ++j) \
		for (i = 0; i < p->nb_col; ++i) { \
			if (Case(p,_i,_j) == E_Pion) { \
				sign[ndx/8] |= x; \
			} \
			if (!(x >>= 1)) { \
				ikey += sign[ndx/8]; \
				ikey += (ikey << 10); \
				ikey ^= (ikey >> 6); \
				x = 0x1 << 7; \
			} \
			++ndx; \
		} \
	if (x != (0x1 << 7)) \
	{ \
		--ndx; \
		ikey += sign[ndx/8]; \
		ikey += (ikey << 10); \
		ikey ^= (ikey >> 6); \
	} \
	ikey += (ikey << 3); \
	ikey ^= (ikey >> 11); \
	ikey += (ikey << 15); \
	*key = ikey & (s->hash_size - 1); \
	return sign; \
}

SIGN(_,i,j)
SIGN(V,p->nb_col-i-1,j)
SIGN(H,i,p->nb_lig-j-1)
SIGN(S,p->nb_lig-j-1,p->nb_col-i-1)
SIGN(B,j,i)



/*
 * Returns 
 * 	1   if found in failed hash
 *      < 0 if error
 */
static int AlreadyFailed(solution_t *s, uint32_t *key, uint8_t **newsign)
{
	int sz = s->sign_sz;
	unsigned long int i = 0;
	uint8_t *sign;
	int found = 0;
	if (!(sign = Calc_Sign(s, key)))
		return -1;
	*newsign = sign;
	for (i = 0; i < s->hash[*key].nb_items; ++i) {
		if (memcmp(s->hash[*key].items[i].sign, sign, sz) == 0) {
			s->hash[*key].items[i].hit = 1;
			found = 1;
			break;
		}
	}
	if (found) 
		free(sign);
	return found;
}

static int AddFailedSign(solution_t *s, int pass, uint32_t key, uint8_t *sign)
{
	struct hash *hash = &s->hash[key];
	if (hash->nb_items+1 >= hash->size) {
		if (!hash->size)
			hash->size = 1;
		else 
			hash->size <<= 1;
		if (hash->size > s->max_hash_size)
			return_error3(1,"Max size %u reached for pass=%d, key=%x",
				     s->max_hash_size,pass,key);
		REALLOC(hash->size, hash->items, 2, "Hash[%d][%x]", pass, key);
	}
	hash->items[hash->nb_items].sign = sign;
	hash->items[hash->nb_items].hit = 0;
	++hash->nb_items;
	return 0;
}

static int AddFailed(solution_t *s, int pass, uint32_t key, uint8_t *sign)
{
	int rc;
	if ((rc = AddFailedSign(s, pass, key, sign)))
		return rc;
#define TESTSYM(letter) \
	do { \
		if (s->letter##_sym) { \
			if (!(sign = Calc##letter##Sign(s, &key)))\
 				return -2; \
			if ((rc = AddFailedSign(s, pass, key, sign))) \
				return rc; \
		} \
	} while (0)
	TESTSYM(H);
	TESTSYM(V);
	TESTSYM(S);
	TESTSYM(B);
	return 0;
}


static int FreeSolution(solution_t *s)
{
	unsigned int i,j;
	for (i = 0; i < s->hash_size; ++i) {
		for (j = 0; j < s->hash[i].nb_items; ++j) {
			free(s->hash[i].items[j].sign);
		}
	}
	FreePlateau(s->plateau);
	free(s->hash);
	free(s->random);
	free(s->liste);
	return 0;
}

static int NewSolution(solution_t *s, plateau_t *p, int random, unsigned int max_hash_size, uint32_t hash_size, int stats)
{
	unsigned int i,j;
	size_t max = MAX(p->nb_lig,p->nb_col);
	int nb_pion = p->nb_col*p->nb_lig;
	/* Set up solution variables */
	s->nb_pion = nb_pion;
	s->plateau = p;
	s->stats = stats;
	s->hash_size = 1 << hash_size;
	MALLOC(p->nb_pion, s->liste, 1, "Liste des coups");
	s->sign_sz = nb_pion / 8;
	s->max_hash_size = max_hash_size;
	if (nb_pion % 8)	/* FIXME: if no pion at end... */
		s->sign_sz++;
	/* Alloc hashes */
	info2("Using hash of %d bits (%.3g entries)", hash_size, (float)s->hash_size);
	MALLOC(s->hash_size,s->hash,-1, " hash");
	for (j = 0; j < s->hash_size; ++j) {
		s->hash[j].nb_items = 0;
		s->hash[j].size = 0;
		s->hash[j].items = NULL;
	}
	/* Look for symetries */
#define LOOKSYM(letter, string, new_i, new_j) \
	do { \
 		s->letter##_sym = 1; \
		for (j = 0; j < p->nb_lig; ++j) \
			for (i = 0; i < p->nb_col; ++i) \
				if (Case(p,i,j) != Case(p,new_i,new_j)) \
					s->letter##_sym = 0; \
		if (s->letter##_sym) \
			info(string " symetry found"); \
	} while (0)
	LOOKSYM(H, "Vertical      (|)", p->nb_col-i-1, j);
	LOOKSYM(V, "Horizontal    (-)", i, p->nb_lig-j-1);
	if (p->nb_lig == p->nb_col) {
		LOOKSYM(S, "Diagonal      (/)", p->nb_lig-j-1, p->nb_col-i-1);
		LOOKSYM(B, "Anti-diagonal (\\)", j, i);
	} else {
		s->S_sym = 0;
		s->B_sym = 0;
	}

	/* Alloc and Feed random */
	srand(time(NULL));
	MALLOC(max,s->random,-1,"Random array");
	for (i = 0; i < max; ++i) {
		unsigned int val;
		while (1) {
			int found = 0;
			val = rand() % max;
			for (j = 0; j < i; ++j)
				if (s->random[j] == val)
					found = 1;
			if (!found)
				break;
		}
		s->random[i] = (random ? val : i);
		if (random)
			info2("Random:%d->%d",i,s->random[i]);
	}

	if (s->stats >= 1)
		MALLOC(s->hash_size, s->array, -1, "stat array");
	else 
		s->array = NULL;
	return 0;
}




/*
 * Hash stats
 */
#define issup(i,j) (i > j)

int compareitems(const void *a, const void *b)
{
	const struct stat_item *ia = (struct stat_item *)a;
	const struct stat_item *ib = (struct stat_item *)b;
	return issup(ia->value, ib->value);
}


void StatHash(solution_t *s)
{
	unsigned int key,i;
	unsigned int hit = 0;
	int sz = s->sign_sz;
	compteur_t total_sum = 0;
	compteur_t max = 0;
	unsigned int nb_empty = 0;
	unsigned int maxi = 0;
	struct hash *hash = s->hash;
	/* Sort hash into list */
	for (key = 0; key < s->hash_size; ++key) {
		for (i = 0; i < hash[key].nb_items; ++i)  {
			if (hash[key].items[i].hit)
				hit++;
		}
		total_sum += hash[key].nb_items;
		if (issup(hash[key].nb_items, max)) {
			max = hash[key].nb_items;
			maxi = key;
		}
		if (!issup(hash[key].nb_items, 0))
			nb_empty++;
		if (s->stats >= 2) {
			s->array[i].key = key;
			s->array[i].value = hash[key].nb_items;
		}
	}
	if (!total_sum)
		return ;
	info4("Hash: usage: %02.2f%%, hit: %02.2f%%, max: %lu, total entries: %lu",
	      ((double)(s->hash_size-nb_empty))/s->hash_size*100, 
	      ((double)(hit))/(total_sum)*100,
	      max,
	      total_sum);
	if (s->stats < 2)
		return ;

	/* Sort array */
	qsort(s->array, s->hash_size, sizeof(*s->array), &compareitems);

	/* Display signatures for max */
	for (i = 0; i < hash[maxi].nb_items; ++i) {
		DisplaySign(hash[maxi].items[i].sign, sz, maxi);
	}
	return ;
}





/*
 *******************************************************************************
 * Plateau
 */
plateau_t *NewPlateau(void)
{
	plateau_t *p;
	MALLOC(1, p, NULL, "Plateau");
	p->nb_col = 0;
	p->nb_lig = 0;
	p->x_fin = -1;
	p->y_fin = -1;
	p->nb_pion = 0;
	p->grille = NULL;
	return p;
}

static int ResizePlateau(plateau_t *p)
{
	unsigned int min_i = 0;
	unsigned int min_j = 0;
	int max_i = 0;
	int max_j = 0;
	unsigned int i,j;
	/* min_j */
	for (min_j = 0; min_j < p->nb_lig; ++min_j) {
		for (i = 0; i < p->nb_col; ++i) 
			if (Case(p,i,min_j) != E_Interdit)
				break;
		if (i != p->nb_col)
			break;
	}
	/* max_j */
	for (max_j = p->nb_lig-1; max_j >= 0; --max_j) {
		for (i = 0; i < p->nb_col; ++i) 
			if (Case(p,i,max_j) != E_Interdit)
				break;
		if (i != p->nb_col)
			break;
	}
	/* min_i */
	for (min_i = 0; min_i < p->nb_col; ++min_i) {
		for (j = 0; j < p->nb_lig; ++j) 
			if (Case(p,min_i,j) != E_Interdit)
				break;
		if (j != p->nb_lig)
			break;
	}
	/* max_i */
	for (max_i = p->nb_col-1; max_i >= 0; --max_i) {
		for (j = 0; j < p->nb_lig; ++j) 
			if (Case(p,max_i,j) != E_Interdit)
				break;
		if (j != p->nb_lig)
			break;
	}
	if (min_j != 0)
		info1("Found real plateau starting at j = %d", min_j);
	if (max_j >= 0 && max_j != (int)p->nb_lig-1)
		info1("Found real plateau ending at j = %d", max_j);
	if (min_i != 0)
		info1("Found real pleteau starting at i = %d", min_i);
	if (max_i >= 0 && max_i != (int)p->nb_col-1)
		info1("Found real plateau ending at i = %d",max_i);
	/* debug("FIXME: compute new plateau !"); */
	return 0;
}

plateau_t *DupPlateau(plateau_t *p)
{
	plateau_t *new;
	unsigned int i;
	size_t sz;
	if (!(new = NewPlateau()))
		return NULL;
	new->nb_col = p->nb_col;
	new->nb_lig = p->nb_lig;
	new->x_fin = p->x_fin;
	new->y_fin = p->y_fin;
	new->nb_pion = p->nb_pion;
	sz = new->nb_col*new->nb_lig;
	MALLOC(sz,new->grille,NULL,"grille");
	for (i = 0; i < sz; ++i)
		new->grille[i] = p->grille[i];
	return new;
}

int ReallouePlateau(plateau_t *p, size_t nb_col)
{
	size_t size;
	unsigned int i,j;
	p->nb_lig++;
	if (nb_col > p->nb_col) {
		/* Realloue un tableau avec la nouvelle largeur et 
		 * copie l'ancien tableau dans le nouveau (et complete avec 
		 * l'etat interdit 
		 * Note: on aurait pu juste 'reallouer' la grille courante
		 * mais le deplacement aurait ete plus complique (partir de la 
		 * fin)
		 * */
		etat_t *grille2;
		size = (p->nb_lig+1)*nb_col;
		/*debug("Increase size %d-> %zd",p->nb_col, nb_col);*/
		MALLOC(size, grille2, 1, "Grille");
		for(j = 0; j < p->nb_lig-1; ++j) {
			for(i = 0; i < p->nb_col; ++i)
				CaseGrille(grille2,nb_col,i,j) = Case(p,i,j);
			for(i = p->nb_col; i < nb_col; ++i)
				CaseGrille(grille2,nb_col,i,j) = E_Interdit;
		}
		free(p->grille);
		p->grille = grille2;
		p->nb_col = nb_col;
	} else {
		/* Ajoute juste une ligne */
		size = p->nb_lig*p->nb_col;
		/*debug("Realloc %zd", size);*/
		if (!(p->grille = realloc(p->grille, size*sizeof(*p->grille))))
			return_error1(1,"Cannot reallocate %u",(unsigned int)size);
	}
	for(i = 0; i < p->nb_col; ++i)
		Case(p,i,p->nb_lig-1) = E_Interdit;
	return 0;
}

/* FIXME: use an internal buffer */
int DisplayPlateau(const plateau_t *p)
{
	unsigned int i,j;
	for (j = 0; j < p->nb_lig; ++j) {
		if (j == 0) {
			printf("X/Y ");
			for (i = 0; i < p->nb_col; ++i) 
				printf("%d",i%10);
			printf(" (%u) \n", p->nb_pion);
		}
		printf("%02d: ",j);
		for (i = 0; i < p->nb_col; ++i) {
			switch (Case(p,i,j)) {
			case E_Interdit:
				printf(" ");
				break;
			case E_Vide:
				if (i == (unsigned int)p->x_fin && 
				    j == (unsigned int)p->y_fin)
					printf(".");
				else
					printf("_");
				break;
			case E_Pion:
				printf("x");
				break;
			default:
				printf("Z");
				break;
			}
		}
		printf("\n");
	}
	return 0;
}



int AjouteLigne(const char *buffer, size_t s, plateau_t *p)
{
	int rc;
	unsigned int i;
	size_t real_size = s;
	unsigned int real_i;
	/*debug("AjouteLigne(%.*s,%zd)",(int)s,buffer,s);*/
	for (i = 0; i < s; ++i)
		if (buffer[i] != 'x' &&
		    buffer[i] != '.' &&
		    buffer[i] != ' ')
			real_size--;
	if ((rc = ReallouePlateau(p, real_size)))
		return rc;
	for (i = 0, real_i = 0;  i < s; ++i) {
		switch (buffer[i]) {
		case 'x':
			Case(p,real_i,p->nb_lig-1) = E_Pion;
			real_i++;
			p->nb_pion++;
			break;
		case '.':
			if (p->x_fin >= 0 /* && p->y_fin >= 0 */)
				return_error4(1, "Duplicate end. Already @%d:%d,"
						" and now @%d:%d", 
						p->x_fin, p->y_fin,
						real_i, p->nb_lig-1);
			Case(p,real_i,p->nb_lig-1) = E_Vide;
			p->x_fin = real_i;
			p->y_fin = p->nb_lig-1;
			real_i++;
			break;
		case ' ':
			real_i++;
			break;
		default:
			/* Ignore */
			break;
		}
	}
	return 0;
}

plateau_t *ReadFile(const char *file)
{
	char *buffer;
	FILE *f;
	unsigned int s = DEFAULT_SIZE;
	size_t len = 0;
	plateau_t *p;
	int offset = 0;
	int nb_ligne = 0;
	if (!(f = fopen(file, "r"))) {
		fprintf(stderr, "Error: Cannot open `%s': ", file);
		perror("");
		return NULL;
	}
	if (!(p = NewPlateau()))
		return NULL;
	MALLOC(s,buffer,NULL,"Input buffer");
	while (fgets(buffer+offset, s-offset, f)) {
		len = strlen(buffer);
		if (len == s-1 && buffer[len-1] != '\n') {
			/* Buffer is too short */
			long int s2 = s << 1;
			if (s2 < s)
				return_error1(NULL, "Max size reached %u",s);
			s = s2;
			if (!(buffer = realloc(buffer, s*sizeof(*buffer))))
				return_error1(NULL, "Cannot allocate %u",s);
			offset = len;
		} else {
			/* Full line */
			++nb_ligne;
			if (buffer[len-1] == '\n')
				len--;
			if (AjouteLigne(buffer, len, p))
				return NULL;
			offset = 0;
		}
	}
	if (!feof(f))
		return_error1(NULL, "Reading `%s' failed", file);
	if (offset && AjouteLigne(buffer, len, p))
		return NULL;
	if (p->x_fin <= 0 /* || p->y_fin<= 0 */)
		return_error(NULL, "No end found !");
	fclose(f);
	free(buffer);
	return p;
}

void FreePlateau(plateau_t *p)
{
	free(p->grille);
	free(p);
	return;
}



/*
 *******************************************************************************
 * Resolution
 */
#define Ok(move,p,x,y) \
	move##move(p,x,y) == E_Vide && \
	move(p,x,y) == E_Pion 
#define Apply(move,s,x,y) \
	do { \
		move##move(s->plateau,x,y) = E_Pion; \
		move(s->plateau,x,y) = E_Vide; \
		Case(s->plateau,x,y) = E_Vide; \
		s->plateau->nb_pion--; \
	} while (0)
#define TryNext(move,s,i,j) \
	do { \
		int rc; \
		try = 1; \
		Apply(move,s,i,j); \
		s->liste[pass].x = i;\
		s->liste[pass].y = j;\
		s->liste[pass].s = S_##move; \
		if ((rc = Solve(s, pass+1)) <= 0) \
		{ \
			free(sign); \
			return rc; \
		} \
		s->plateau->nb_pion++; \
		move##move(s->plateau,i,j) = E_Vide; \
		move(s->plateau,i,j) = E_Pion; \
		Case(s->plateau,i,j) = E_Pion; \
	} while (0)
#define Try(cond,move,s,p,i,j) \
	do { \
		if (cond && Ok(move,p,i,j)) \
			TryNext(move,s,i,j); \
	} while (0)



static int Solve(solution_t *s, int pass)
{
	unsigned int i,j;
	uint8_t *sign = NULL;
	uint32_t key = 0;
	plateau_t *p = s->plateau;
	int rc;
	int try = 0;
	if (REMEMBER(pass,s->nb_pion) && 
	    (rc = AlreadyFailed(s, &key, &sign)))
		return rc;
	for (j = 0; j < p->nb_lig; ++j) {
		for (i = 0; i < p->nb_col; ++i) {
			int ri = s->random[i];
			int rj = s->random[j];
			if (Case(p,ri,rj) == E_Pion) {
				/* FIXME: optimize order ? 
				 * weigth toward final peg ?
				 */
				Try(rj > 1,Nord,s,p,ri,rj);
				Try(ri > 1,Ouest,s,p,ri,rj);
				Try(rj < (int)p->nb_lig-2,Sud,s,p,ri,rj);
				Try(ri < (int)p->nb_col-2,Est,s,p,ri,rj);
			}
		}
	}
	if (p->nb_pion == 1 && Case(p,p->x_fin,p->y_fin) == E_Pion) {
		/* YES ! */
		free(sign);
		return 0;
	}
	if (REMEMBER(pass,s->nb_pion)) {
		if (!try) /* terminal => no queuing ? easy next step ?*/
			free(sign);
		else if (AddFailed(s,pass,key,sign) != 0)
			return -1;
	}
	return 1;
}

static int ShowSolution(const solution_t *s, int display)
{
	unsigned int i;
	unsigned int nb_pass = s->plateau->nb_pion-1;
	if (display)
		DisplayPlateau(s->plateau);
	for (i = 0; i < nb_pass; ++i) {
		printf("[%u]%d:%d -> %s\n", i, s->liste[i].y, s->liste[i].x, Name(s->liste[i].s));
#		define DoIt(move,s,i) \
			do { \
				if (s->liste[i].s == S_##move) \
					Apply(move,s,s->liste[i].x, \
					      	     s->liste[i].y);\
			} while (0)
		DoIt(Nord,s,i);
		DoIt(Sud,s,i);
		DoIt(Est,s,i);
		DoIt(Ouest,s,i);
		if (display)
			DisplayPlateau(s->plateau);
	}
	return 0;
}



/*
 *******************************************************************************
 *
 */

#define USAGE \
	"Usage: solitaire [options]\n" \
	"Options:\n" \
	" -f <file>   : read <file> \n" \
	" -d          : display solution\n" \
	" -s          : display hash stats at the end. Twice to show overlap\n" \
	" -S <nb>     : nb bits for hash (default: 18)\n" \
	" -M <nb>     : nb max of items per key (default: 500)\n" \
	" -r          : use random for peg order\n" \
        " -h          : display this help\n" \
	" -v          : display version\n" 


int main(int argc, char *argv[])
{
	const char *file = NULL;
	char *end;
	char opt;
	plateau_t *p, *copy;
	solution_t s;
	int display = 0;
	int stats = 0;
	int random = 0;
	unsigned int max_hash_size = 500;
	uint32_t hash_size = 18;
	int rc = 1;

	while ((opt = getopt(argc, argv, "hvf:dsrS:M:")) != -1) {
		switch (opt) {
		case 'f':
			if (!file)
				file = optarg;
			else
				return_error1(1, "Already given file `%s'",
					     file);
			break;
		case 'h':
			fprintf(stdout, USAGE);
			return 0;
			break;
		case 'v':
			fprintf(stdout, "solitaire version 0.1\n");
			return 0;
			break;
		case 'd':
			display = 1;
			break;
		case 'S': 
			hash_size = strtoul(optarg,&end,10);
			if (*end != '\0') {
				fprintf(stderr, "Invalid -S <nb>: %s\n" USAGE,
					optarg);
				return 1;
			}
			break;
		case 'M':
			max_hash_size = strtoul(optarg,&end,10);
			if (*end != '\0') {
				fprintf(stderr, "Invalid -M <size>: %s\n" USAGE,
					optarg);
				return 1;
			}
			break;
		case 's':
			if (!++stats)
				return_error(1, "Too many -s ;D");
			break;
		case 'r':
			if (!++random)
				return_error(1, "Too many -r ;D");
			break;
		case '?': 
			fprintf(stderr,USAGE);
			return 1;
			break;
		}
	}
	if (!file) {
		fprintf(stderr, "No file given\n" USAGE);
		return 1;
	}


	/* Plateau */
	if (!(p = ReadFile(file)))
		return 1;
	if (ResizePlateau(p))
		return 1;
	if (!(copy = DupPlateau(p)))
		return 1;
	/* Solution */
	if (NewSolution(&s, p, random, max_hash_size, hash_size, stats))
		return 1;
	g_sol = &s;

	/* Let's go */
	oldint = signal(SIGINT, handler);
	oldtstp = signal(SIGTSTP, handler);
	DisplayPlateau(p);
	if (Solve(&s, 0) == 0) {
		s.plateau = copy;
		ShowSolution(&s,display);
		if (stats)
			StatHash(&s);
		rc = 0;
	} else {
		fprintf(stderr, "Could not find solution in `%s'\n",file);
	}
	signal(SIGINT, oldint);
	signal(SIGTSTP, oldtstp);
	/* Freeing */
	FreeSolution(&s);
	FreePlateau(rc == 0 ? p:copy);
	return rc;
}
